%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% FI:PA171 - Digital Filters
%
% topic: Image restoration
% authors: David Svoboda, Marek Kasik 
%          Masaryk University (c) 2019
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% read an image
lena = imread('lena_std.tif');

% create some PSF
disk = double(imread('disk.tif'));

% normalize psf
disk=disk/sum(sum(disk));

% create the image of 'noise'
strength = 32.0;
noise=(rand(512,512)-0.5)*strength;

% blur 'lena'
lena_blur=abs(fftshift(ifft2(fft2(lena).*fft2(disk))));

% add noise the blurred 'lena'
lena_blur_noise=lena_blur+noise;

% show the degraded image
imagesc(lena_blur_noise);
colormap(gray);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Unconstrained (inverse filtering)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% transfer function 
fH=fft2(disk);

% inverse filtering 
result=fftshift(ifft2(fft2(lena_blur_noise)./fH));

% show the result
imagesc(abs(result));
colormap(gray);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Wiener filtering + inverse filtering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% power spectra of 'lena_blur_noise'
Pg=fft2(lena_blur_noise).*fft2(flipdim(flipdim(lena_blur_noise,1),2));

% power spectra of 'noise image'
Pn=fft2(noise).*fft2(flipdim(flipdim(noise,1),2));

% Wiener filtr for noise reduction
F=(Pg-Pn)./Pg;

% inverse filter
D=fft2(disk);

% Wiener filter & inverse filtering
result=fftshift(ifft2(fft2(lena_blur_noise).*(F./D)));

% show the result
imagesc(abs(result));
colormap(gray);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Constrained
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% matrix Q (Laplace)
Q=zeros(512,512);
Q(256,256)=1.0;
Q(255,256)=-0.25;
Q(257,256)=-0.25;
Q(256,255)=-0.25;
Q(256,257)=-0.25;

Ht=fft2(flipdim(flipdim(disk,1),2));
HtH=fft2(disk).*fft2(flipdim(flipdim(disk,1),2));
QtQ=fft2(Q).*fft2(flipdim(flipdim(Q,1),2));
lambda=1.0;

% deconvolution filter
D=lambda*Ht./(lambda*HtH+QtQ);

% the application of deconvolution filter
result=fftshift(ifft2(fft2(lena_blur_noise).*D));

% show the results
imagesc(abs(result));
colormap(gray);
